﻿using System;
using System.Collections.Generic;
using System.Numerics;
using System.Drawing;
using System.Diagnostics;
using System.Windows.Forms;
using System.Windows;

namespace WindowsFormsApp1
{
    public partial class Form1 : Form
    {
        State state;
        Stopwatch stopwatch;
        long lastTicks;
	long remainingTicks;
        bool trackMouse;

        public Form1()
        {
            InitializeComponent();

            this.stopwatch = new Stopwatch();
            this.stopwatch.Start();
	    this.lastTicks = this.stopwatch.ElapsedTicks;
            this.remainingTicks = 0;
            this.state = new State(this.pictureBox1.Width, this.pictureBox1.Height);
            this.trackMouse = false;

            this.timer1.Start();
        }

        private void pictureBox1_Paint(object sender, PaintEventArgs e)
        {
            this.state.Draw(e.Graphics);
        }

        private void timer1_Tick(object sender, EventArgs e)
        {
            if (this.trackMouse)
            {
                System.Drawing.Point mousePointer = this.pictureBox1.PointToClient(Cursor.Position);
                this.state.SetAttractor(new Vector2(mousePointer.X, mousePointer.Y));
            }

            // If we want 5 ms resolution, we need 200 ticks per second, i.e. frequency / 200 ticks resolution.
            float resolutionMillis = 10.0f;
            float ticksPerResolution = Stopwatch.Frequency * resolutionMillis / 1000.0f;

            long ticks = this.stopwatch.ElapsedTicks;
            long elapsedTicks = (ticks - this.lastTicks) + this.remainingTicks;
            this.lastTicks = ticks;
            this.remainingTicks = (long)(elapsedTicks % ticksPerResolution);

            int updateCount = (int)Math.Floor(elapsedTicks / ticksPerResolution);

//             System.Console.WriteLine("Timer ticks per second: {0}", Stopwatch.Frequency);
//             System.Console.WriteLine("Timer resolution: {0}", ticksPerResolution);
//             System.Console.WriteLine("Elapsed ticks: {0}", elapsedTicks);
//             System.Console.WriteLine("Performing {0} updates", updateCount);
//             System.Console.WriteLine("Remaining time: {0}", this.remainingTicks);

            for (int i = 0; i < updateCount; i++) {
//                 System.Console.Write(".");
                this.state.Update(resolutionMillis);
            }

//             System.Console.WriteLine();

            this.pictureBox1.Invalidate();
        }

        private void pictureBox1_MouseDown(object sender, MouseEventArgs e)
        {
            this.trackMouse = true;
        }

        private void pictureBox1_MouseUp(object sender, MouseEventArgs e)
        {
            this.trackMouse = false;
        }

        private void pictureBox1_MouseLeave(object sender, EventArgs e)
        {
            this.trackMouse = false;
        }
    }

    class State
    {
        List<Item> items;
        Vector2 gravity;
        int width, height;

        public State(int width, int height)
        {
            this.width = width;
            this.height = height;

            this.items = new List<Item>();

            this.gravity = new Vector2(0.0f, 1.0f);

            items.Add(new Ball(new Vector2(100.0f), new Vector2(0.0f), 20, Pens.Red));

            // Bottom.
            items.Add(new Wall(new Vector2(0.0f, 300.0f), 500.0f, 20.0f, Brushes.Black));

            // Mid-air blocks.
            items.Add(new Wall(new Vector2(80.0f, 200.0f), 20.0f, 20.0f, Brushes.BlueViolet));
            items.Add(new Wall(new Vector2(150.0f, 200.0f), 20.0f, 20.0f, Brushes.BlueViolet));

            // Left.
            items.Add(new Wall(new Vector2(0.0f, 100.0f), 20.0f, 200.0f, Brushes.Black));

            // Right.
            items.Add(new Wall(new Vector2(480f, 100.0f), 20.0f, 200.0f, Brushes.Black));
        }

        public void Update(float time)
        {
            foreach (Item i in this.items)
            {
                i.Push(this.gravity * (time / 100.0f));
            }

            for (int i = 0; i < this.items.Count - 1; i++)
            {
                for (int j = i + 1; j < this.items.Count; j++)
                {
                    this.items[i].SolveCollision(this.items[j]);
                }
            }

            foreach (Item i in this.items)
            {
                i.Update();
            }

            //foreach (Item i in this.items)
            //{
            //    i.position = new Vector2((i.position.X + this.width) % this.width, (i.position.Y + this.height) % this.height);
            //}
        }

        public void Draw(Graphics g)
        {
            foreach (Item i in this.items)
            {
                i.Draw(g);
            }
        }

        public void SetAttractor(Vector2 pos)
        {
            foreach (Item i in this.items)
            {
                Ball b = i as Ball;
                if (b != null)
                {
                    Vector2 vecToMouse = pos - b.position;
                    b.Push(vecToMouse / 100.0f);
                }
            }
        }
    }

    abstract class Item
    {
        public Vector2 position;
        protected Vector2 velocity;

        protected Item(Vector2 position, Vector2 velocity)
        {
            this.position = position;
            this.velocity = velocity;
        }
        virtual public void Update()
        {
            this.velocity *= 0.99f;
            this.position += velocity;
        }
        abstract public void Draw(Graphics g);
        virtual public void Push(Vector2 force)
        {
            this.velocity += force;
        }

        abstract public void SolveCollision(Item other);

        abstract public System.Windows.Rect GetBoundingBox();

        public static void CollideBallBottom(Ball b, Wall w)
        {
            System.Windows.Rect ballBoundingBox = b.GetBoundingBox();
            System.Windows.Rect wallBoundingBox = w.GetBoundingBox();

            System.Windows.Rect intersection = Rect.Intersect(ballBoundingBox, wallBoundingBox);
            if (!intersection.IsEmpty)
            {
                Vector2 intersection_middle = new Vector2((float)(intersection.Left + intersection.Right) / 2.0f,
                    (float)(intersection.Top + intersection.Bottom) / 2.0f);
                Vector2 collisionVec = Vector2.Normalize(b.position - intersection_middle);

                double collisionAngle = Math.Acos(Vector2.Dot(collisionVec, Vector2.Normalize(b.velocity)));
                if (collisionAngle > Math.PI/2.0f)
                {
                    // Actual collision.

                    Vector2 velocityAfterCollision = Vector2.Reflect(b.velocity, collisionVec);
                    b.velocity = velocityAfterCollision * b.elasticity;
                }
            }
        }
    }

    class Ball : Item
    {
        Pen p;
        float size;
        public float elasticity;

        public Ball(Vector2 position, Vector2 velocity, float size, Pen p) : base(position, velocity)
        {
            this.size = size;
            this.p = p;
            this.elasticity = 0.8f;
        }
        public override void Draw(Graphics g)
        {
            g.DrawEllipse(this.p, this.position.X - this.size/2.0f, this.position.Y - this.size / 2.0f, this.size, this.size);
        }

        public override System.Windows.Rect GetBoundingBox()
        {
            return new Rect(this.position.X - this.size / 2.0f, this.position.Y - this.size / 2.0f, this.size, this.size);
        }

        public override void SolveCollision(Item other)
        {
            Ball b = other as Ball;
            if (b != null)
            {
                throw new NotImplementedException();
            }

            Wall w = other as Wall;
            if (w != null)
            {
                Item.CollideBallBottom(this, w);
            }
        }
    }

    class Wall : Item
    {
        Brush b;
        float width;
        float height;

        public Wall(Vector2 position, float width, float height, Brush b): base(position, new Vector2(0.0f))
        {
            this.b = b;
            this.width = width;
            this.height = height;
        }
        public override void Push(Vector2 force)
        {
            // Do nothing.
            return;
        }
        public override void Draw(Graphics g)
        {
            g.FillRectangle(this.b, this.position.X, this.position.Y, this.width, this.height);
        }

        public override void SolveCollision(Item other)
        {
            Ball b = other as Ball;
            if (b != null)
            {
                Item.CollideBallBottom(b, this);
            }
        }

        public override System.Windows.Rect GetBoundingBox()
        {
            return new Rect(this.position.X, this.position.Y, this.width, this.height);
        }
    }
}
